package net.sourceforge.iriverter;

import java.io.*;
import java.util.*;

public class Converter extends Thread {
	private List jobs, notSplitVideos;
	private ProgressDialogInfo progressDialogInfo;
	private ConverterOptions converterOptions;
	private Process proc;
	private boolean isCanceled;
	private int exitCode;
	
	public Converter(List jobs, ProgressDialogInfo progressDialogInfo, ConverterOptions converterOptions) {
		this.jobs = Converter.checkForOverwritingFiles(Converter.expandSingleJobsToMultiple(Converter.removeInvalidJobs(jobs), converterOptions));
		this.progressDialogInfo = progressDialogInfo;
		this.converterOptions = converterOptions;
		isCanceled = false;
		
		notSplitVideos = new ArrayList();
		
		progressDialogInfo.setTotalJobs(this.jobs.size());
	}
	
	public static List removeInvalidJobs(List jobs) {
		List newJobs = new ArrayList();
		
		for (int i = 0; i < jobs.size(); i++)
			if (jobs.get(i) instanceof SingleVideoInfo) {
				SingleVideoInfo singleVideoInfo = (SingleVideoInfo) jobs.get(i);
				if (new File(singleVideoInfo.getInputVideo()).exists() && !singleVideoInfo.getOutputVideo().equals(""))
					newJobs.add(singleVideoInfo);
			} else if (jobs.get(i) instanceof DirectoryInfo) {
				DirectoryInfo directoryInfo = (DirectoryInfo) jobs.get(i);
				if (new File(directoryInfo.getInputDirectory()).exists() && !directoryInfo.getOutputDirectory().equals(""))
					newJobs.add(directoryInfo);
			} else if (jobs.get(i) instanceof DVDInfo) {
				DVDInfo dvdInfo = (DVDInfo) jobs.get(i);
				if (!dvdInfo.getDrive().equals("") && !dvdInfo.getOutputVideo().equals(""))
					newJobs.add(dvdInfo);
			} else if (jobs.get(i) instanceof ManualSplitInfo) {
				ManualSplitInfo manualSplitInfo = (ManualSplitInfo) jobs.get(i);
				if (!manualSplitInfo.getVideo().equals("") && manualSplitInfo.getMarks().length > 2)
					newJobs.add(manualSplitInfo);
			}
			
		return newJobs;
	}
	
	public static List expandSingleJobsToMultiple(List jobs, ConverterOptions converterOptions) {
		List newJobs = new ArrayList();
		
		for (int i = 0; i < jobs.size(); i++)	
			if (jobs.get(i) instanceof DirectoryInfo)
				newJobs.addAll(convertDirectoryToSingleVideos((DirectoryInfo) jobs.get(i), converterOptions));
			else if (jobs.get(i) instanceof DVDInfo)
				newJobs.addAll(separateDVDChaptersToSingleDVDJobs((DVDInfo) jobs.get(i)));
			else if (jobs.get(i) instanceof ManualSplitInfo)
				newJobs.addAll(separateMultipleSplitJobsToOneSplitJob((ManualSplitInfo) jobs.get(i)));
			else
				newJobs.add(jobs.get(i));
		
		return newJobs;
	}
	
	public static List convertDirectoryToSingleVideos(DirectoryInfo directoryInfo, ConverterOptions converterOptions) {
		List newJobs = new ArrayList();
		
		String[] directory = new File(directoryInfo.getInputDirectory()).list(new VideoFileFilter());
		
		for (int i = 0; i < directory.length; i++)
			if (new File(directoryInfo.getInputDirectory() + File.separator + directory[i]).isDirectory() && directoryInfo.getConvertSubdirectories())
				newJobs.addAll(convertDirectoryToSingleVideos(new DirectoryAdapter(directoryInfo.getInputDirectory() + File.separator + directory[i], directoryInfo.getOutputDirectory() + File.separator + directory[i], directoryInfo.getConvertSubdirectories()), converterOptions));
			else if (new File(directoryInfo.getInputDirectory() + File.separator + directory[i]).isFile())
				newJobs.add(new SingleVideoAdapter(directoryInfo.getInputDirectory() + File.separator + directory[i], directoryInfo.getOutputDirectory() + File.separator + directory[i].substring(0, directory[i].lastIndexOf('.')) + "." + converterOptions.getShortDevice() + ".avi"));
			
		return newJobs;
	}
	
	public static List separateDVDChaptersToSingleDVDJobs(DVDInfo dvdInfo) {
		List newJobs = new ArrayList();
	
		Chapters[] chapters = dvdInfo.getChapters();
		
		if (chapters == null)
			newJobs.add(dvdInfo);
		else
			for (int i = 0; i < chapters.length; i++) {
				String outputVideo = "";
				if (chapters[i].getFirstChapter() == chapters[i].getLastChapter())
					outputVideo = dvdInfo.getOutputVideo().substring(0, dvdInfo.getOutputVideo().lastIndexOf('.')) + ".ch" + chapters[i].getFirstChapterPadded() + ".avi";
				else
					outputVideo = dvdInfo.getOutputVideo().substring(0, dvdInfo.getOutputVideo().lastIndexOf('.')) + ".ch" + chapters[i].getFirstChapterPadded() + "-" + chapters[i].getLastChapterPadded() + ".avi";
				
				newJobs.add(new DVDAdapter(dvdInfo.getDrive(), dvdInfo.getTitle(), new Chapters[]{chapters[i]}, dvdInfo.getAudioStream(), dvdInfo.getSubtitles(), outputVideo));
			}
		
		return newJobs;
	}
	
	public static List separateMultipleSplitJobsToOneSplitJob(ManualSplitInfo manualSplitInfo) {
		List newJobs = new ArrayList();
		
		for (int i = 0; (i + 1) < manualSplitInfo.getMarks().length; i++)
			newJobs.add(new ManualSplitAdapter(manualSplitInfo.getVideo(), new String[]{manualSplitInfo.getMarks()[i], manualSplitInfo.getMarks()[i + 1]}, i + 1));
		
		return newJobs;		
	}
	
	public static List checkForOverwritingFiles(List jobs) {
		List newJobs = new ArrayList();
		
		for (int i = 0; i < jobs.size(); i++) {
			if (!(jobs.get(i) instanceof OutputVideoInfo))
				newJobs.add(jobs.get(i));
			else if (new File(((OutputVideoInfo) jobs.get(i)).getOutputVideo()).exists()) {
				if (OverwriteDialog.overwriteFile(((OutputVideoInfo) jobs.get(i)).getOutputVideo()))
					newJobs.add(jobs.get(i));
			}
			else
				newJobs.add(jobs.get(i));
		}
		
		return newJobs;
	}
	
	public void run() {
		for (int i = 0; i < jobs.size() && !isCanceled; i++) {
			progressDialogInfo.setCurrentJob(i + 1);
			
			if (jobs.get(i) instanceof SingleVideoInfo)
				convertSingleVideo((SingleVideoInfo) jobs.get(i));;
			if (jobs.get(i) instanceof DVDInfo)
				convertDVD((DVDInfo) jobs.get(i));
			if (jobs.get(i) instanceof ManualSplitInfo)
				manuallySplitVideo((ManualSplitInfo) jobs.get(i));
		}
		
		if (!isCanceled)
			progressDialogInfo.complete(exitCode == 0);
	}
	
	public void cancel() {
		isCanceled = true;
		
		if (proc != null)
			proc.destroy();
	}
	
	public void convertSingleVideo(SingleVideoInfo singleVideoInfo) {
		progressDialogInfo.setInputVideo(new File(singleVideoInfo.getInputVideo()).getName());
		progressDialogInfo.setOutputVideo(new File(singleVideoInfo.getOutputVideo()).getName());
		progressDialogInfo.setStatus("Gathering information about the input video...");
		
		MPlayerInfo info = new MPlayerInfo(singleVideoInfo.getInputVideo(), null);
		
		int scaledWidth = converterOptions.getWidth();
		int scaledHeight = (info.getHeight() * converterOptions.getWidth()) / info.getWidth();
		
		if (scaledHeight > converterOptions.getHeight()) {
			scaledWidth = (scaledWidth * converterOptions.getHeight()) / scaledHeight;
			scaledHeight = converterOptions.getHeight();
		}
		
		String vf = "harddup";
		if (converterOptions.getPanAndScan())
			vf = "scale=" + ((int) ((info.getWidth()) * (((double) converterOptions.getHeight()) / (double) info.getHeight()))) + ":" + converterOptions.getHeight() + ",crop=" + converterOptions.getWidth() + ":" + converterOptions.getHeight() + "," + vf;
		else
			vf = "scale=" + scaledWidth + ":" + scaledHeight + ",expand=" + converterOptions.getWidth() + ":" + converterOptions.getHeight() + "," + vf;
		
		String af = "resample=44100";
		if (converterOptions.getNormalizeVolume())
			af = "volnorm," + af;
		if (!converterOptions.getAutoSync()) {
			int offset = converterOptions.getAudioDelay();
			if (offset <= 0)
				af = "advance=" + -offset + "," + af;
			else
				af = "delay=" + offset + ":" + offset + "," + af;
		}
		
		List commandList = new ArrayList();
		
		commandList.add(MPlayerInfo.getMPlayerPath() + "mencoder");
		commandList.add(singleVideoInfo.getInputVideo());
		commandList.add("-o");
		commandList.add(singleVideoInfo.getOutputVideo());
		commandList.add("-ovc");
		commandList.add("xvid");
		commandList.add("-xvidencopts");
		commandList.add("bitrate=" + converterOptions.getVideoBitrate() + ":max_bframes=0");
		commandList.add("-oac");
		commandList.add("mp3lame");
		commandList.add("-lameopts");
		commandList.add("mode=2:cbr:br=" + converterOptions.getAudioBitrate());
		commandList.add("-vf");
		commandList.add(vf);
		commandList.add("-af");
		commandList.add(af);
		commandList.add("-ofps");
		commandList.add("" + converterOptions.getFrameRate());
		commandList.add("-srate");
		commandList.add("44100");
		
		if (!converterOptions.getAutoSync()) {
			commandList.add("-mc");
			commandList.add("0");
		}
		
		commandList.addAll(extraOptionsToTokenList(converterOptions));
		
		String[] command = new String[commandList.size()];
		for (int i = 0; i < command.length; i++)
			command[i] = (String) commandList.get(i);
		
		if (!isCanceled) {
			new File(singleVideoInfo.getOutputVideo()).getParentFile().mkdirs();
			splitVideo(singleVideoInfo.getOutputVideo(), runConversionCommand(command));
		}
	}
	
	public void convertDVD(DVDInfo dvdInfo) {
		String inputVideo = "Title " + dvdInfo.getTitle() + " of the DVD at " + dvdInfo.getDrive();
		
		Chapters[] chapters = dvdInfo.getChapters();
		if (chapters != null) {
			if (chapters[0].getFirstChapter() == chapters[0].getLastChapter())
				inputVideo = "Chapter " + chapters[0].getFirstChapter() + " of " + inputVideo;
			else 
				inputVideo = "Chapters " + chapters[0].getFirstChapter() + "-" + chapters[0].getLastChapter() + " of " + inputVideo;
		}
		
		progressDialogInfo.setInputVideo(inputVideo);
		progressDialogInfo.setOutputVideo(new File(dvdInfo.getOutputVideo()).getName());
		progressDialogInfo.setStatus("Gathering information about the input video...");
		
		MPlayerInfo info = new MPlayerInfo("dvd://" + dvdInfo.getTitle(), dvdInfo.getDrive());
		
		int scaledWidth = converterOptions.getWidth();
		int scaledHeight = (info.getHeight() * converterOptions.getWidth()) / info.getWidth();
		
		if (scaledHeight > converterOptions.getHeight()) {
			scaledWidth = (scaledWidth * converterOptions.getHeight()) / scaledHeight;
			scaledHeight = converterOptions.getHeight();
		}
		
		String vf = "filmdint=io=" + ((int) (info.getFrameRate() * 100)) + ":" + (converterOptions.getFrameRate() * 100);
		if (converterOptions.getPanAndScan())
			vf += ",scale=" + ((int) ((info.getWidth()) * (((double) converterOptions.getHeight()) / (double) info.getHeight()))) + ":" + converterOptions.getHeight() + ",crop=" + converterOptions.getWidth() + ":" + converterOptions.getHeight();
		else
			vf += ",scale=" + scaledWidth + ":" + scaledHeight + ",expand=" + converterOptions.getWidth() + ":" + converterOptions.getHeight();
		vf += ",harddup";
		
		String af = "resample=44100";
		if (converterOptions.getNormalizeVolume())
			af = "volnorm," + af;
		if (!converterOptions.getAutoSync()) {
			int offset = converterOptions.getAudioDelay();
			if (offset <= 0)
				af = "advance=" + -offset + "," + af;
			else
				af = "delay=" + offset + ":" + offset + "," + af;
		}
		
		List commandList = new ArrayList();
		
		commandList.add(MPlayerInfo.getMPlayerPath() + "mencoder");
		commandList.add("-dvd-device");
		commandList.add(dvdInfo.getDrive());
		commandList.add("dvd://" + dvdInfo.getTitle());
		commandList.add("-o");
		commandList.add(dvdInfo.getOutputVideo());
		commandList.add("-ovc");
		commandList.add("xvid");
		commandList.add("-xvidencopts");
		commandList.add("bitrate=" + converterOptions.getVideoBitrate() + ":max_bframes=0");
		commandList.add("-oac");
		commandList.add("mp3lame");
		commandList.add("-lameopts");
		commandList.add("mode=2:cbr:br=" + converterOptions.getAudioBitrate());
		commandList.add("-vf");
		commandList.add(vf);
		commandList.add("-af");
		commandList.add(af);
		commandList.add("-ofps");
		commandList.add("" + converterOptions.getFrameRate());
		commandList.add("-srate");
		commandList.add("44100");
		
		if (dvdInfo.getAudioStream() > -1) {
			commandList.add("-aid");
			commandList.add("" + dvdInfo.getAudioStream());
		}
		
		if (dvdInfo.getSubtitles() > -1) {
			commandList.add("-sid");
			commandList.add("" + dvdInfo.getSubtitles());
		}
		
		if (dvdInfo.getChapters() != null) {
			commandList.add("-chapter");
			commandList.add(dvdInfo.getChapters()[0].getFirstChapter() + "-" + dvdInfo.getChapters()[0].getLastChapter());
		}
		
		if (!converterOptions.getAutoSync()) {
			commandList.add("-mc");
			commandList.add("0");
		}
		
		commandList.addAll(extraOptionsToTokenList(converterOptions));
		
		String[] command = new String[commandList.size()];
		for (int i = 0; i < command.length; i++)
			command[i] = (String) commandList.get(i);
		
		if (!isCanceled) {
			new File(dvdInfo.getOutputVideo()).getParentFile().mkdirs();
			splitVideo(dvdInfo.getOutputVideo(), runConversionCommand(command));
		}
	}
	
	public void manuallySplitVideo(ManualSplitInfo manualSplitInfo) {
		String outputVideo = manualSplitInfo.getVideo().substring(0, manualSplitInfo.getVideo().lastIndexOf('.')) + ".part" + manualSplitInfo.getPart() + ".avi";
		
		progressDialogInfo.setInputVideo(manualSplitInfo.getVideo());
		progressDialogInfo.setOutputVideo(outputVideo);
		
		runConversionCommand(new String[]{MPlayerInfo.getMPlayerPath() + "mencoder", manualSplitInfo.getVideo(), "-o", outputVideo, "-ovc", "copy", "-oac", "copy", "-ss", manualSplitInfo.getMarks()[0], "-endpos", manualSplitInfo.getMarks()[1]});
	}
	
	public int runConversionCommand(String[] command) {
		MencoderStreamParser inputStream = null;
		MencoderStreamParser errorStream = null;
		
		System.out.print("DEBUG--- Running: ");
		for (int i = 0; i < command.length; i++)
			System.out.print(command[i] + " ");
		System.out.println();
		
		try {
			proc = Runtime.getRuntime().exec(command);
			
			inputStream = new MencoderStreamParser(progressDialogInfo);
			inputStream.parse(new BufferedReader(new InputStreamReader(proc.getInputStream())));
			errorStream = new MencoderStreamParser(progressDialogInfo);
			errorStream.parse(new BufferedReader(new InputStreamReader(proc.getErrorStream())));
			
			exitCode = proc.waitFor();
			proc = null;
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		if (exitCode != 0)
			isCanceled = true;
		
		return isCanceled ? 0 : ((errorStream.getLength() > -1) ? errorStream.getLength() : inputStream.getLength());
	}
	
	public void splitVideo(String inputVideo, int length) {
		if (!converterOptions.getDevice().equals("H300 Series") || length < 3600)
			return;
		
		if (!converterOptions.getAutomaticallySplit()) {
			notSplitVideos.add(inputVideo);
			return;
		}
		
		int pieces = (length / 3600) + 1;
		for (int i = 0; i < pieces; i++) {
			String outputVideo = inputVideo.substring(0, inputVideo.lastIndexOf('.')) + ".part" + (i + 1) + ".avi";
			
			progressDialogInfo.setInputVideo(new File(inputVideo).getName());
			progressDialogInfo.setOutputVideo(new File(outputVideo).getName());
			
			runConversionCommand(new String[]{MPlayerInfo.getMPlayerPath() + "mencoder", inputVideo, "-o", outputVideo, "-ovc", "copy", "-oac", "copy", "-ss", "" + (length / pieces) * i, "-endpos", "" + (length / pieces)});
		}
	}
	
	public List getNotSplitVideos() {
		return notSplitVideos;
	}
	
	public static List extraOptionsToTokenList(ConverterOptions converterOptions) {
		List extraOptionsTokens = new ArrayList();
		
		String extraOptions = converterOptions.getExtraOptions();
		StringTokenizer tokens = new StringTokenizer(extraOptions);
		while (tokens.hasMoreTokens())
			extraOptionsTokens.add(tokens.nextToken());
		
		return extraOptionsTokens;
	}
}
